@@ -1,4 +1,6 @@
 import collections
+import collections.abc
+import functools
 import os.path
 import re
 import unittest
@@ -45,7 +47,8 @@ class Changelog(list):
     _ignore_re = re.compile(r'^(?:  |\s*\n)')
 
     class Entry(object):
-        __slot__ = 'distribution', 'source', 'version', 'urgency', 'maintainer', 'date'
+        __slot__ = ('distribution', 'source', 'version', 'urgency',
+                    'maintainer', 'date')
 
         def __init__(self, **kwargs):
             for key, value in kwargs.items():
@@ -57,7 +60,8 @@ class Changelog(list):
         if file:
             self._parse(version, file)
         else:
-            with open(os.path.join(dir, "debian/changelog"), encoding="UTF-8") as f:
+            with open(os.path.join(dir, "debian/changelog"),
+                      encoding="UTF-8") as f:
                 self._parse(version, f)
 
     def _parse(self, version, f):
@@ -72,7 +76,8 @@ class Changelog(list):
             elif top_match is None:
                 top_match = self._top_re.match(line)
                 if not top_match:
-                    raise Exception('invalid top line %d in changelog' % line_no)
+                    raise Exception('invalid top line %d in changelog' %
+                                    line_no)
                 try:
                     v = version(top_match.group('version'))
                 except Exception:
@@ -82,16 +87,19 @@ class Changelog(list):
             else:
                 bottom_match = self._bottom_re.match(line)
                 if not bottom_match:
-                    raise Exception('invalid bottom line %d in changelog' % line_no)
+                    raise Exception('invalid bottom line %d in changelog' %
+                                    line_no)
 
-                self.append(self.Entry(distribution=top_match.group('distribution'),
-                                       source=top_match.group('source'),
-                                       version=v,
-                                       urgency=top_match.group('urgency'),
-                                       maintainer=bottom_match.group('maintainer'),
-                                       date=bottom_match.group('date')))
+                self.append(self.Entry(
+                    distribution=top_match.group('distribution'),
+                    source=top_match.group('source'),
+                    version=v,
+                    urgency=top_match.group('urgency'),
+                    maintainer=bottom_match.group('maintainer'),
+                    date=bottom_match.group('date')))
                 top_match = bottom_match = None
 
+
 class Version(object):
     _epoch_re = re.compile(r'\d+$')
     _upstream_re = re.compile(r'[0-9][A-Za-z0-9.+\-:~]*$')
@@ -110,9 +118,9 @@ class Version(object):
             upstream, revision = rest, None
         else:
             upstream, revision = rest[0:split], rest[split+1:]
-        if ((epoch is not None and not self._epoch_re.match(epoch)) or
-            not self._upstream_re.match(upstream) or
-            (revision is not None and not self._revision_re.match(revision))):
+        if (epoch is not None and not self._epoch_re.match(epoch)) or \
+           not self._upstream_re.match(upstream) or \
+           (revision is not None and not self._revision_re.match(revision)):
             raise RuntimeError(u"Invalid debian version")
         self.epoch = epoch and int(epoch)
         self.upstream = upstream
@@ -136,7 +144,8 @@ class Version(object):
     @property
     def debian(self):
         from warnings import warn
-        warn(u"debian argument was replaced by revision", DeprecationWarning, stacklevel=2)
+        warn(u"debian argument was replaced by revision", DeprecationWarning,
+             stacklevel=2)
         return self.revision
 
 
@@ -188,29 +197,29 @@ class _VersionTest(unittest.TestCase):
 
     def test_invalid_epoch(self):
         with self.assertRaises(RuntimeError):
-            v = Version('a:1')
+            Version('a:1')
         with self.assertRaises(RuntimeError):
-            v = Version('-1:1')
+            Version('-1:1')
         with self.assertRaises(RuntimeError):
-            v = Version('1a:1')
+            Version('1a:1')
 
     def test_invalid_upstream(self):
         with self.assertRaises(RuntimeError):
-            v = Version('1_2')
+            Version('1_2')
         with self.assertRaises(RuntimeError):
-            v = Version('1/2')
+            Version('1/2')
         with self.assertRaises(RuntimeError):
-            v = Version('a1')
+            Version('a1')
         with self.assertRaises(RuntimeError):
-            v = Version('1 2')
+            Version('1 2')
 
     def test_invalid_revision(self):
         with self.assertRaises(RuntimeError):
-            v = Version('1-2_3')
+            Version('1-2_3')
         with self.assertRaises(RuntimeError):
-            v = Version('1-2/3')
+            Version('1-2/3')
         with self.assertRaises(RuntimeError):
-            v = Version('1-2:3')
+            Version('1-2:3')
 
 
 class VersionLinux(Version):
@@ -379,7 +388,7 @@ class _VersionLinuxTest(unittest.TestCase):
         self.assertFalse(v.linux_revision_other)
 
     def test_other_revision(self):
-        v = VersionLinux('4.16.5-1+revert+crng+ready') # from #898087
+        v = VersionLinux('4.16.5-1+revert+crng+ready')  # from #898087
         self.assertFalse(v.linux_revision_experimental)
         self.assertFalse(v.linux_revision_security)
         self.assertFalse(v.linux_revision_backports)
@@ -393,7 +402,7 @@ class _VersionLinuxTest(unittest.TestCase):
         self.assertTrue(v.linux_revision_other)
 
 
-class PackageArchitecture(collections.MutableSet):
+class PackageArchitecture(collections.abc.MutableSet):
     __slots__ = '_data'
 
     def __init__(self, value=None):
@@ -421,7 +430,7 @@ class PackageArchitecture(collections.MutableSet):
 
     def extend(self, value):
         if isinstance(value, str):
-            for i in re.split('\s', value.strip()):
+            for i in re.split(r'\s', value.strip()):
                 self.add(i)
         else:
             raise RuntimeError
@@ -493,7 +502,7 @@ class PackageRelation(list):
 
     def extend(self, value, override_arches=None):
         if isinstance(value, str):
-            value = (j.strip() for j in re.split(',', value.strip()))
+            value = (j.strip() for j in re.split(r',', value.strip()))
         for i in value:
             self.append(i, override_arches)
 
@@ -529,7 +538,7 @@ class PackageRelationGroup(list):
 
     def extend(self, value, override_arches=None):
         if isinstance(value, str):
-            value = (j.strip() for j in re.split('\|', value.strip()))
+            value = (j.strip() for j in re.split(r'\|', value.strip()))
         for i in value:
             self.append(i, override_arches)
 
@@ -537,7 +546,8 @@ class PackageRelationGroup(list):
 class PackageRelationEntry(object):
     __slots__ = "name", "operator", "version", "arches", "restrictions"
 
-    _re = re.compile(r'^(\S+)(?: \((<<|<=|=|!=|>=|>>)\s*([^)]+)\))?(?: \[([^]]+)\])?(?: <([^>]+)>)?$')
+    _re = re.compile(r'^(\S+)(?: \((<<|<=|=|!=|>=|>>)\s*([^)]+)\))?'
+                     r'(?: \[([^]]+)\])?((?: <[^>]+>)*)$')
 
     class _operator(object):
         OP_LT = 1
@@ -573,7 +583,8 @@ class PackageRelationEntry(object):
             self._op = self.operators[value]
 
         def __neg__(self):
-            return self.__class__(self.operators_text[self.operators_neg[self._op]])
+            return self.__class__(
+                self.operators_text[self.operators_neg[self._op]])
 
         def __str__(self):
             return self.operators_text[self._op]
@@ -597,7 +608,7 @@ class PackageRelationEntry(object):
         if self.arches:
             ret.extend((' [', ' '.join(self.arches), ']'))
         if self.restrictions:
-            ret.extend((' <', ' '.join(self.restrictions), '>'))
+            ret.extend((' ', str(self.restrictions)))
         return ''.join(ret)
 
     def parse(self, value):
@@ -612,13 +623,103 @@ class PackageRelationEntry(object):
             self.operator = None
         self.version = match[2]
         if match[3] is not None:
-            self.arches = re.split('\s+', match[3])
+            self.arches = re.split(r'\s+', match[3])
         else:
             self.arches = []
-        if match[4] is not None:
-            self.restrictions = re.split('\s+', match[4])
+        self.restrictions = PackageBuildRestrictFormula(match[4])
+
+
+class PackageBuildRestrictFormula(set):
+    _re = re.compile(r' *<([^>]+)>(?: +|$)')
+
+    def __init__(self, value=None):
+        if value:
+            self.update(value)
+
+    def __str__(self):
+        return ' '.join(f'<{i}>' for i in sorted(self))
+
+    def add(self, value):
+        if isinstance(value, str):
+            value = PackageBuildRestrictList(value)
+        elif not isinstance(value, PackageBuildRestrictList):
+            raise ValueError("got %s" % type(value))
+        super(PackageBuildRestrictFormula, self).add(value)
+
+    def update(self, value):
+        if isinstance(value, str):
+            pos = 0
+            for match in self._re.finditer(value):
+                if match.start() != pos:
+                    break
+                pos = match.end()
+            if pos != len(value):
+                raise ValueError(f'invalid restriction formula "{value}"')
+            value = (match.group(1) for match in self._re.finditer(value))
+        for i in value:
+            self.add(i)
+
+    # TODO: union etc.
+
+
+class PackageBuildRestrictList(frozenset):
+    # values are established in frozenset.__new__ not __init__, so we
+    # implement __new__ as well
+    def __new__(cls, value=()):
+        if isinstance(value, str):
+            if not re.fullmatch(r'[^()\[\]<>,]+', value):
+                raise ValueError(f'invalid restriction list "{value}"')
+            value = (PackageBuildRestrictTerm(i) for i in value.split())
         else:
-            self.restrictions = []
+            for i in value:
+                if not isinstance(i, PackageBuildRestrictTerm):
+                    raise ValueError
+        return super(PackageBuildRestrictList, cls).__new__(cls, value)
+
+    def __str__(self):
+        return ' '.join(str(i) for i in sorted(self))
+
+    # TODO: union etc.
+
+
+@functools.total_ordering
+class PackageBuildRestrictTerm(object):
+    def __init__(self, value):
+        if not isinstance(value, str):
+            raise ValueError
+        match = re.fullmatch(r'(!?)([^()\[\]<>,!\s]+)', value)
+        if not match:
+            raise ValueError(f'invalid restriction term "{value}"')
+        self.negated = bool(match.group(1))
+        self.profile = match.group(2)
+
+    def __str__(self):
+        return ('!' if self.negated else '') + self.profile
+
+    def __eq__(self, other):
+        return (self.negated == other.negated
+                and self.profile == other.profile)
+
+    def __lt__(self, other):
+        return (self.profile < other.profile
+                or (self.profile == other.profile
+                    and not self.negated and other.negated))
+
+    def __hash__(self):
+        return hash(self.profile) ^ int(self.negated)
+
+
+def restriction_requires_profile(form, profile):
+    # An empty restriction formula does not require any profile.
+    # Otherwise, a profile is required if each restriction list
+    # includes it without negation.
+    if len(form) == 0:
+        return False
+    term = PackageBuildRestrictTerm(profile)
+    for lst in form:
+        if term not in lst:
+            return False
+    return True
 
 
 class _ControlFileDict(dict):
@@ -671,6 +772,7 @@ class Package(_ControlFileDict):
         ('Breaks', PackageRelation),
         ('Conflicts', PackageRelation),
         ('Description', PackageDescription),
+        ('Build-Profiles', PackageBuildRestrictFormula),
     ))
 
 
